//     ======================= RcgExample.cpp ===============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 14.04.2008
//     **********************************************************************

#include <stdio.h>
#include "..\..\include\AssociativeMemory.h"
#include "highgui.h"

#define cAMFileName "am.dat"

void doSearchObject(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void* args)
{
	static char ObjName[10][64] = { "Hand-open", "Hand-close",
		"Book", "Post card", "Body spray", "Toy \"Dog\"", "Toy \"Pig\"",
		"Clock", "Hand", "Test area"};
	char OutNumStr[256];

	int i;
	static double bTime,eTime,Tf;
	CvFont font;
	cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );

	IplImage* pImg = (IplImage*)args;

	switch(aState) {
		case cProcessingIsBegun:
			bTime = (double)cvGetTickCount();
			Tf = (double)cvGetTickFrequency()*1000.0;
			break;

		case cUnknownObject:
/*
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(rand() % 255,rand() % 255,rand() % 255),3,8,0);
			printf(">U< Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);
			cvShowImage("Result", pImg);
			cvWaitKey(10);
*/
			break;

		case cRecognizedObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x, apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,128,0),1,8,0);
			printf("R   Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);
			break;

		case cGeneralizedObject:
		case cTrackedObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,0,0),2,8,0);

			sprintf(OutNumStr,"%s",ObjName[*apData-1]);
			cvPutText(pImg, OutNumStr, cvPoint(apParam->ObjRect.x+10, apParam->ObjRect.y +
				                               apParam->ObjRect.height+24), &font, CV_RGB(255,255,0));

			printf("G-T Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);

			if(apParam->ppUserTrackingInfo) {
				#define cTrcLen 100
				struct TTrcInfo {
					CvPoint TrcPoint[cTrcLen];
					long Age;
				};
				if(!*apParam->ppUserTrackingInfo) {
					*apParam->ppUserTrackingInfo = new TTrcInfo;
					memset(*apParam->ppUserTrackingInfo, 0, sizeof(TTrcInfo));
				} else {
					TTrcInfo* pTrcInfo = (TTrcInfo*)*apParam->ppUserTrackingInfo;
					memcpy(&pTrcInfo->TrcPoint[0], &pTrcInfo->TrcPoint[1], sizeof(CvPoint)*(cTrcLen-1));
					pTrcInfo->TrcPoint[cTrcLen-1] = cvPoint(apParam->ObjRect.x + (apParam->ObjRect.width>>1),
						                                    apParam->ObjRect.y + (apParam->ObjRect.height>>1));
					if(pTrcInfo->Age < cTrcLen) { pTrcInfo->Age++; }

					if(pTrcInfo->Age > 2) {
						for(i=1; i<pTrcInfo->Age; i++) {
							cvLine(pImg, pTrcInfo->TrcPoint[cTrcLen-i-1],
								         pTrcInfo->TrcPoint[cTrcLen-i], cvScalar(255,255,0),2,8,0);
						}
						cvCircle(pImg, pTrcInfo->TrcPoint[cTrcLen-1],4,cvScalar(0,255,0),2,8,0);
					}
				}
			}
			break;

		case cLearnThisObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x +  apParam->ObjRect.width,
								       apParam->ObjRect.y +  apParam->ObjRect.height),
									   cvScalar(255,255,0),2,8,0);
			printf("L   Obj: [%s] ...... Similarity: %7.4f, Bases: %lu\n",
				   ObjName[*apData-1], apParam->Similarity, apParam->TotalABases);
			break;

		case cProcessingIsFinished:
			eTime = (double)cvGetTickCount();
			sprintf(OutNumStr,"%5.1f ms",(eTime-bTime)/Tf);
			cvPutText(pImg, OutNumStr, cvPoint(0, 24), &font, CV_RGB(0,255,0));
			break;
	}
}

int main(int argc, char* argv[])
{
	int       i;
	ULONG     FilmLengthMs  = 0;
	double    FilmFrameRate;
	
	IplImage  *OutImg, *HLS_Img,*HueImg, *GrayImg, *xMaskImg, *MaskImg;
	CvHistogram* HueHist = NULL;

	CvAssociativeMemory32S am;

	
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== Recognition help ===\n\n"
			   "RecognitionEDV.exe [file.avi | 0..9] [/r ]\n"
			   " 0..9 - capture device number\n"
			   " /r   - write result in file_out.avi\n");
		return 0;
	}

	char* argv1 = _strlwr(argv[1]);
	if(!(strstr(argv1,".avi") || isdigit(argv[1][0]))){
		printf("AVI files are processing only!\n\n");
		return -1;
	}

	#define cStrMax 256
	char FileName[cStrMax];
	char OutStr[cStrMax];
	char CommandsStr[cStrMax];
	memset(CommandsStr,0,cStrMax);
	for(i=2; i<argc; i++) strcat(CommandsStr,argv[i]);
	memset(FileName,0,cStrMax);
	CvCapture*     InAVI;
    IplImage*      frame = 0;
	CvVideoWriter* OutAVI;

	bool WriteResultToFile = false;
	if(strstr(CommandsStr,"/r"))  WriteResultToFile = true;
	int WIDTH, HEIGHT;
	CvSize ImgSize;

    if(isdigit(argv[1][0])) {
		printf("Processing capture...\n");
        InAVI = cvCaptureFromCAM( argc == 2 ? argv[1][0] - '0' : 0 );
		if( !InAVI ) {
			fprintf(stderr,"Could not initialize capturing...\n");
			return -1;
		}
		frame = cvQueryFrame(InAVI);
		ImgSize = cvGetSize(frame);
		WIDTH   = ImgSize.width;
		HEIGHT  = ImgSize.height;
		FilmFrameRate = 12;
    } else {
		printf("Processing avi file...\n");
		memcpy(FileName,argv1,strstr(argv1,".avi")-argv1);
		InAVI = cvCaptureFromFile(argv1);
		if(!InAVI) {
			printf("Can't find file %s.avi\n",FileName);
			return -1;
		}
		cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,1);
		FilmLengthMs  = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
		cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,0);
		WIDTH         = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_WIDTH));
		HEIGHT        = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_HEIGHT));
		FilmFrameRate = cvGetCaptureProperty(InAVI,CV_CAP_PROP_FPS);
		ImgSize       = cvSize(WIDTH, HEIGHT);
    }

	OutImg   = cvCreateImage( ImgSize, 8, 3 );
	HLS_Img  = cvCreateImage( ImgSize, 8, 3 );
	HueImg   = cvCreateImage( ImgSize, 8, 1 );
	GrayImg  = cvCreateImage( ImgSize, 8, 1 );
	xMaskImg = cvCreateImage( ImgSize, 8, 1 );
	MaskImg  = cvCreateImage( ImgSize, 8, 1 );
	
	if(WriteResultToFile) {
		sprintf(OutStr,"%s_out.avi",FileName);
		OutAVI = cvCreateVideoWriter(OutStr, CV_FOURCC('c','v','i','d'), FilmFrameRate, cvSize(WIDTH,HEIGHT), 1 );
		if(!OutAVI) {
			OutAVI = cvCreateVideoWriter(OutStr, -1, FilmFrameRate, cvSize(WIDTH,HEIGHT), 1 );
		}
	}

	uint64 ABIndex;
	uint64 ABHitCounter;

	CvRect InterestArea;
	InterestArea.x = 390;
	InterestArea.y = 250;
	InterestArea.width  = 220;
	InterestArea.height = 220;

	if(am.Load(cAMFileName)) {
		printf("\n  Load associative bases from file: %s\n"
		   "Associative bases ......... %-7lu\n"
		   "Associative Levels ........ %-7u\n"
		   "Current base index ........ %-7I64u\n"
		   "Wr/Rd counter ............. %-7I64u\n\n",
		   cAMFileName, am.GetTotalABases(), am.GetTotalLevels(), am.GetCurIndex(), am.GetWrRdCounter());
	} else { am.Create(cvSize(InterestArea.width, InterestArea.height),3); }

	cvNamedWindow("Result", 0);

	long PLoop        = 0;
	long TrainingLoop = 400;

	for(;;) {
		
		frame = cvQueryFrame(InAVI);
		if( !frame ) break;

		cvFlip(frame);

		cvCopy(frame, OutImg);
		
		cvCvtColor(frame, GrayImg, CV_BGR2GRAY);

		if(PLoop) {
				cvRectangle( OutImg, cvPoint(InterestArea.x,   InterestArea.y),
								 cvPoint(InterestArea.x + InterestArea.width,
								 InterestArea.y + InterestArea.height), cvScalar(0,255,0),2,8,0);
			PLoop--;
		} else {

			cvCvtColor(frame, HLS_Img, CV_BGR2HSV);

			cvInRangeS( HLS_Img, cvScalar(0,30,10,0), cvScalar(180,256,256,0), xMaskImg);
			cvCvtPixToPlane(HLS_Img, HueImg, 0, 0, 0);

			if(!HueHist) {
				int HistDims = 16;
				static float HueRanges_arr[] = {0,180};
				float* HueRanges = HueRanges_arr;
				HueHist = cvCreateHist( 1, &HistDims, CV_HIST_ARRAY, &HueRanges, 1 );
				cvCalcHist(&HueImg, HueHist, 0, xMaskImg);
				float max_val = 0;
				cvGetMinMaxHistValue(HueHist, 0, &max_val, 0, 0);
				cvConvertScale(HueHist->bins, HueHist->bins, (max_val)?(255.0/max_val):0, 0);
			}
			
			cvCalcBackProject(&HueImg, MaskImg, HueHist);
			cvAnd(MaskImg, xMaskImg, MaskImg, 0);
			cvNot(MaskImg, MaskImg);

			cvNamedWindow("Mask", 0);
			cvShowImage("Mask", MaskImg);
				
			am.SetImage(GrayImg, MaskImg);

			if(TrainingLoop) {

				long Data = 1;
				CvRect SAr;
				am.Write(InterestArea, &Data);
				double Similarity;
				am.Read(InterestArea, &SAr, 0, &ABIndex, &ABHitCounter, &Similarity, true); 
				printf("Hitting! Index = %I64u, HitCounter = %I64u, TotalABases = %lu, Similarity = %f\n",
				   ABIndex, ABHitCounter, am.GetTotalABases(), Similarity);
				cvRectangle( OutImg, cvPoint(InterestArea.x,   InterestArea.y),
								 cvPoint(InterestArea.x + InterestArea.width,
								 InterestArea.y + InterestArea.height), cvScalar(0,0,255),2,8,0);
				TrainingLoop--;
			} else {
				am.ObjectTracking(doSearchObject, OutImg);
			}
		}

		ULONG POS_MSEC = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
		int hour        = int(POS_MSEC/(3600000));
		POS_MSEC        %= 3600000L;
		int minute      = int(POS_MSEC/(60000));
		POS_MSEC        %= 60000L;
		int second      = int(POS_MSEC/1000);
		int millisecond = int(POS_MSEC % 1000);
		printf("==> film current position: %2.2d:%2.2d:%2.2d.%3.3d\n",
				hour, minute, second, millisecond);

	    cvShowImage("Result", OutImg);

		if(WriteResultToFile) { cvWriteFrame(OutAVI, OutImg); }

		if( cvWaitKey(10) >= 0 ) break;
	}
	
	am.Save(cAMFileName);

	if(WriteResultToFile) { cvReleaseVideoWriter(&OutAVI); }
	cvReleaseCapture(&InAVI);

	cvReleaseImage(&OutImg);
	cvReleaseImage(&HLS_Img);
	cvReleaseImage(&HueImg);
	cvReleaseImage(&GrayImg);
	cvReleaseImage(&xMaskImg);
	cvReleaseImage(&MaskImg);
	
	return 0;
}
